import java.util.Arrays;
import java.util.LinkedList;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.RandomAccessFile;
/** 
 * SnakeOil (c) 2015 - 2019
 * https://bitbucket.org/MysticLord/
 */
public class EDC_ECC_Generator
{
    private long[] EDC_crctable;
    private long[][] L2sq;
    private long L2_RAW = 0x800;    // 2048
    private long L2_P = 43*2*2;     // 172
    private long L2_Q = 26*2*2;     // 104

    /**
     * Create an object of type EDC_ECC_Generator
     */
    public EDC_ECC_Generator()
    {
        EDC_crctable = genEDCcrctable();
        L2sq = genL2sq();
    }

    public static void main(String[] args)
    {
        EDC_ECC_Generator generator = new EDC_ECC_Generator();
    }

    /**
     * Read EDC CRC table from text file.
     * @param   long[] tempArray
     */
    private long[] genEDCcrctable()
    {
        LinkedList<Long> tempList = new LinkedList<Long>();
        long[] tempArray = null;
        BufferedReader br = null;
        try
        {
            br = new BufferedReader(new FileReader(new File("EDC_crctable.txt")));
            String line = br.readLine();
            while (line != null)
            {
                tempList.add(Long.decode(line));
                line = br.readLine();
            }
            tempArray = new long[(tempList.size())];
            int counter = 0;
            for (Long curr : tempList)
            {
                tempArray[counter] = (long)curr;
                counter ++;
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
        finally
        {
            if (br != null)
            {
                try
                {
                    br.close();
                }
                catch (Exception e) {}
            }
            return tempArray;
        }
    }

    /**
     * Read the L2sq table from a text file.
     * @return  long[][] tempArray
     */
    private long[][] genL2sq()
    {
        LinkedList<Long[]> tempList = new LinkedList<Long[]>();
        long[][] tempArray = null;
        BufferedReader br = null;
        try
        {
            br = new BufferedReader(new FileReader(new File("L2sq.txt")));
            String line = br.readLine();
            int columnLength = line.split(",").length;
            int rowLength = 0;
            while (line != null)
            {
                String[] lineStringArray = line.split(",");
                Long[] lineLongArray = new Long[lineStringArray.length];
                int i = 0;
                for (String part : lineStringArray)
                {
                    lineLongArray[i] = Long.decode(lineStringArray[i]);
                    i ++;
                }
                tempList.add(lineLongArray);
                line = br.readLine();
                rowLength ++;
            }

            tempArray = new long[rowLength][columnLength];
            for (int rowCounter = 0; rowCounter < rowLength; rowCounter ++)
            {
                for (int columnCounter = 0; columnCounter < columnLength; columnCounter ++)
                {
                    if (columnCounter == 0)
                    {
                        System.out.print("");
                    }
                    tempArray[rowCounter][columnCounter] = tempList.get(rowCounter)[columnCounter];
                }
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
        finally
        {
            if (br != null)
            {
                try
                {
                    br.close();
                }
                catch (Exception e) {}
            }
            return tempArray;
        }
    }

    /**
     * Generate the EDC CRC.  Requires a long[], first index, and last 
     * index on which to operate.  Returns a long that is the CRC32 code.
     * You still need to assign it in parts to 4 ints that will be 
     * written as unsigned bytes.
     * 
     * @param   long[] sector
     * @param   int firstIndex
     * @param   int lastindex
     * @return  long edc
     */
    private long crc32(long[] sector, int firstIndex, int lastIndex)
    {
        long edc = 0;
        for (int i = firstIndex; i < lastIndex; i++)
        {
            edc = (EDC_crctable[(int)(edc ^ sector[i]) & 255] ^ (edc >> 8));
        }
        return edc;
    }

    /**
     * Generates first part of ECC.  Adds null bytes too.
     * @param   long[2060] sector
     * @return  long[2236) newSector
     */
    private long[] encode_L2_P(long[] sector)
    {
        //L2_RAW =  0x800 = 2048
        //L2_P   = 43*2*2 =  172
        //L2_Q   = 26*2*2 =  104

        long base_index = 0;
        long P_index = 4 + L2_RAW + 4 + 8;      // 4 + 2048 + 4 + 8 = 2064
        long target_size = P_index + L2_P;      // 2064 + 172 = 2236

        // we need to make a new array here of the correct size (2236)
        // long[] sector is only 2060 bytes
        long[] newSector = new long[(int)target_size];
        newSector[0] = newSector[1] = newSector[2] = newSector[3] = 0L; // null bytes
        for (int i = 4; i < (sector.length + 4); i++)
        {
            newSector[i] = sector[i - 4]; // account for null bytes
        }
        assert newSector.length == target_size;    // required array length: 2236

        for (int j = 0; j < 43; j++)
        {
            long a = 0;
            long b = 0;
            long index = base_index;
            for (int i = 19; i < 43; i++)
            {
                assert index < (P_index - 1);
                a = a ^ L2sq[i][(int)newSector[(int)index]];
                b = b ^ L2sq[i][(int)newSector[(int)(index + 1)]];
                index += (2 * 43);
            }
            newSector[(int)P_index] = a >> 8;
            newSector[(int)P_index + (43 * 2)] = a & 0xFF;
            newSector[(int)P_index + 1] = b >> 8;
            newSector[(int)P_index + (43 * 2) + 1] = b & 0xFF;
            base_index += 2;
            P_index += 2;
        }
        return newSector;
    }

    /**
     * Generates second part of ECC.
     * @param   long[] sector
     * @return  long[] newSector
     */
    private long[] encode_L2_Q(long[] sector)
    {
        //L2_RAW =  0x800 = 2048
        //L2_P   = 43*2*2 =  172
        //L2_Q   = 26*2*2 =  104

        long base_index = 0;
        long Q_index = 4 + L2_RAW + 4 + 8 + L2_P;   // 4 + 2048 + 4 + 8 + 172 = 2236
        long MOD_INDEX = Q_index;                   // 2236
        long target_size = Q_index + L2_Q;          // 2236 + 104 = 2340

        long[] newSector = new long[(int)target_size];
        for (int i = 0; i < sector.length; i++)
        {
            newSector[i] = sector[i];
        }

        assert newSector.length == target_size;        // required array length: 2340
        int counter = 0;
        for (int j = 0; j < 26; j++)
        {
            long a = 0;
            long b = 0;
            long index = base_index;
            for (int i = 0; i < 43; i++)
            {
                a = a ^ L2sq[i][(int)newSector[(int)index]];
                b = b ^ L2sq[i][(int)newSector[(int)(index + 1)]];
                index += (2 * 44);
                index = index % MOD_INDEX;
            }
            newSector[(int)Q_index] = a >> 8;
            newSector[(int)Q_index + (26 * 2)] = a & 0xFF;
            newSector[(int)Q_index + 1] = b >> 8;
            newSector[(int)Q_index + (26 * 2) + 1] = b & 0xFF;
            base_index += (2 * 43);
            Q_index += 2;
            counter ++;
        }
        return newSector;
    }

    /**
     * Generates EDC and ECC for a form 1 sector.  Requires a long[] of length 2060.
     * @param   long[] sector
     * @return  long[] sector2340
     */
    public long[] get_edc_ecc(long[] sector)
    {
        long edc = crc32(sector, 0, 2056);
        for (int m = 0; m < 4; m++)
        {
            sector[2056 + m] += (edc & 255L);
            edc >>= 8;
        }
        long[] sector2236 = encode_L2_P(sector);
        long[] sector2340 = encode_L2_Q(sector2236);

        return sector2340;
    }

    /**
     * Generates EDC for a form 2 sector.  Requires a long[].  Length doesn't matter, 
     * but should be 2332 bytes.
     * @param   long[] sector
     * @return  long[] newSector
     */
    public long[] get_edc_form2(long[] sector)
    {
        /*                          -    s -    e
         *    12 bytes synch        -    0 -   11
         *     4 bytes header       -   12 -   15
         *     8 bytes subheader    -   16 -   23
         *  2324 bytes data         -   24 - 2347
         *     4 bytes edc          - 2348 - 2351
         *  2352 bytes total        - 
         *  
         *  so we need an array that is 
         *     8 bytes subheader
         *  2324 bytes data
         *  2332 bytes total... but add 4 at some point because we're adding the EDC
         *     
         */
        long[] newSector = new long[2336];
        for (int i = 0; i < sector.length; i++)
        {
            newSector[i] = sector[i];
        }
        long edc = crc32(newSector, 0, 2332);
        for (int j = 0; j < 4; j++)
        {
            newSector[2332 + j] += (edc & 255L);
            edc >>= 8;
        }
        return newSector;
    }
}